<?php

namespace Drupal\openapi_jsonapi\Plugin\openapi\OpenApiGenerator;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Authentication\AuthenticationCollectorInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\Entity\ConfigEntityInterface;
use Drupal\Core\Config\Entity\ConfigEntityTypeInterface;
use Drupal\Core\Entity\ContentEntityStorageInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Link;
use Drupal\Core\ParamConverter\ParamConverterManagerInterface;
use Drupal\Core\Routing\RouteObjectInterface;
use Drupal\Core\Routing\RouteProviderInterface;
use Drupal\Core\Url;
use Drupal\jsonapi\JsonApiResource\RelationshipData;
use Drupal\jsonapi\JsonApiResource\ResourceObject;
use Drupal\jsonapi\JsonApiResource\ResourceObjectData;
use Drupal\jsonapi\JsonApiResource\TopLevelDataInterface;
use Drupal\jsonapi\JsonApiSpec;
use Drupal\jsonapi\Normalizer\Value\CacheableNormalization;
use Drupal\jsonapi\ResourceType\ResourceType;
use Drupal\jsonapi\ResourceType\ResourceTypeRepositoryInterface;
use Drupal\jsonapi\Routing\Routes;
use Drupal\jsonapi\Routing\Routes as JsonApiRoutes;
use Drupal\openapi\Plugin\openapi\OpenApiGeneratorBase;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\Routing\Route;
use Symfony\Component\Serializer\SerializerInterface;

/**
 * Defines an OpenApi Schema Generator for the JsonApi module.
 *
 * @OpenApiGenerator(
 *   id = "jsonapi",
 *   label = @Translation("JSON:API"),
 * )
 */
class JsonApiGenerator extends OpenApiGeneratorBase {

  const JSON_API_UUID_CONVERTER = 'paramconverter.jsonapi.entity_uuid';

  /**
   * Separator for using in definition id strings.
   *
   * Override the default one to use '--' and match jsonapi.
   *
   * @var string
   */
  static $DEFINITION_SEPARATOR = '--';

  /**
   * List of parameters hat should be filtered out on JSON API Routes.
   *
   * @var string[]
   */
  static $PARAMETERS_FILTER_LIST = [
    JsonApiRoutes::RESOURCE_TYPE_KEY,
  ];

  /**
   * Module Handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  private $moduleHandler;

  /**
   * Parameter Converter Manager.
   *
   * @var \Drupal\Core\ParamConverter\ParamConverterManagerInterface
   */
  private $paramConverterManager;

  /**
   * The entity type bundle info.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  private $entityTypeBundleInfo;

  /**
   * Map of normalized resource names.
   *
   * @var array
   */
  protected $resource_names = [];

  /**
   * Store for resource types, keyed by entity definition key.
   *
   * @var ResourceType[]
   */
  protected $resourceTypes = [];

  /**
   * Stub data storage.
   *
   * @var \SplObjectStorage;
   */
  protected $stubDataStorage;

  /**
   * JsonApiGenerator constructor.
   *
   * @param array $configuration
   *   Plugin configuration.
   * @param string $plugin_id
   *   Unique plugin id.
   * @param array|mixed $plugin_definition
   *   Plugin instance definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Routing\RouteProviderInterface $routing_provider
   *   The routing provider.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $field_manager
   *   The field manager.
   * @param \Symfony\Component\Serializer\SerializerInterface $serializer
   *   The serializer.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The current request stack.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The configuration object factory.
   * @param \Drupal\Core\Authentication\AuthenticationCollectorInterface $authentication_collector
   *   The authentication collector.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler service.
   * @param \Drupal\Core\ParamConverter\ParamConverterManagerInterface $param_converter_manager
   *   The parameter converter manager service.
   * @param \Drupal\jsonapi\ResourceType\ResourceTypeRepositoryInterface $resource_type_repository
   *   The resource type manager.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entity_type_bundle_info
   *   The entity type bundle info.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, RouteProviderInterface $routing_provider, EntityFieldManagerInterface $field_manager, SerializerInterface $serializer, RequestStack $request_stack, ConfigFactoryInterface $config_factory, AuthenticationCollectorInterface $authentication_collector, ModuleHandlerInterface $module_handler, ParamConverterManagerInterface $param_converter_manager, protected ResourceTypeRepositoryInterface $resourceTypeRepository, EntityTypeBundleInfoInterface $entity_type_bundle_info) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $routing_provider, $field_manager, $serializer, $request_stack, $config_factory, $authentication_collector);
    $this->moduleHandler = $module_handler;
    $this->paramConverterManager = $param_converter_manager;
    $this->entityTypeBundleInfo = $entity_type_bundle_info;

    $this->resource_names = static::retrieveResourceNames(
      $this->resourceTypeRepository
    );
    // Remove the disabled resource types from the output.
    $this->options['exclude'] = static::findDisabledMethods(
      $this->resourceTypeRepository,
      $this->resource_names
    );
    $this->stubDataStorage = new \SplObjectStorage();
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('router.route_provider'),
      $container->get('entity_field.manager'),
      $container->get('jsonapi.serializer'),
      $container->get('request_stack'),
      $container->get('config.factory'),
      $container->get('authentication_collector'),
      $container->get('module_handler'),
      $container->get('paramconverter_manager'),
      $container->get('jsonapi.resource_type.repository'),
      $container->get('entity_type.bundle.info')
    );
  }

  /**
   * Retrieve all resource names, including any which have been altered.
   *
   * @param \Drupal\jsonapi\ResourceType\ResourceTypeRepositoryInterface $resource_type_repository
   *   Resource type repository.
   *
   * @return array
   *   Keyed array mapping "[entity type]/[bundle]" to resource name.
   */
  protected static function retrieveResourceNames(ResourceTypeRepositoryInterface $resource_type_repository) {
    $map = [];
    foreach ($resource_type_repository->all() as $resourceType) {
      $map[$resourceType->getEntityTypeId() . '/' . $resourceType->getBundle()] = $resourceType->getTypeName();
    }
    return $map;
  }

  /**
   * Introspects all the JSON API resource types and outputs the disabled ones.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\jsonapi\ResourceType\ResourceTypeRepositoryInterface $resource_type_repository
   *   The resource type manager.
   *
   * @return string[]
   *   A list of resource keys to disable.
   */
  protected static function findDisabledMethods(ResourceTypeRepositoryInterface $resource_type_repository, array $resource_map) {
    $extract_resource_type_id = function (ResourceType $resource_type) use ($resource_map) {
      return $resource_map[$resource_type->getEntityTypeId() . '/' . ($resource_type->getBundle() ?? $resource_type->getEntityTypeId())];
    };
    $filter_disabled = function (ResourceType $resourceType) {
      // If there is an isInternal method and the resource is marked as internal
      // then consider it disabled. If not, then it's enabled.
      return method_exists($resourceType, 'isInternal') && $resourceType->isInternal();
    };
    $all = $resource_type_repository->all();
    $disabled_resources = array_filter($all, $filter_disabled);
    $disabled = array_map($extract_resource_type_id, $disabled_resources);
    return $disabled;
  }

  /**
   * {@inheritdoc}
   */
  public function getBasePath() {
    return parent::getBasePath() . $this->getJsonApiBase();
  }

  /**
   * Determine the base for JSON:API's endpoint routes.
   *
   * @return string
   *   The url prefix used for all jsonapi resource endpoints.
   */
  public function getJsonApiBase() {
    return \Drupal::getContainer()->getParameter('jsonapi.base_path');
  }

  /**
   * {@inheritdoc}
   */
  public function getPathItems() {
    $routes = $this->getJsonApiRoutes();
    $api_paths = [];
    foreach ($routes as $route_name => $route) {
      /** @var \Drupal\jsonapi\ResourceType\ResourceType $resource_type */
      $resource_type = $this->getResourceTypeFromRoute($route_name, $route);
      // @todo Enable config entities.
      if (!$resource_type instanceof ResourceType || $this->entityTypeManager->getDefinition($resource_type->getEntityTypeId())->entityClassImplements(ConfigEntityInterface::class)) {
        continue;
      }
      // Store the resource type so we can access it later, while still using
      // a lot of the base class methods.
      $this->storeResourceType($resource_type);
      if (!$this->includeEntityTypeBundle($resource_type->getEntityTypeId(), $resource_type->getBundle())) {
        continue;
      }
      $api_path = [];
      $methods = array_map('strtolower', $route->getMethods());
      foreach ($methods as $method) {
        // This is the "operation object."
        $path_method = [
          'summary' => $this->getRouteMethodSummary($route, $route_name, $method),
          'description' => $this->getRouteMethodDescription($route, $route_name, $method, $resource_type->getTypeName()),
          'parameters' => $this->getMethodParameters($route, $route_name, $resource_type, $method),
          'tags' => [$this->getBundleTag($resource_type->getEntityTypeId(), $resource_type->getBundle())],
          'responses' => $this->getEntityResponsesJsonApi($resource_type, $method, $route_name, $route),
        ];
        if (in_array($method, ['put', 'patch', 'post'])) {
          $path_method['requestBody'] = $this->getRequestBody($resource_type, $method, $route_name, $route);
        }
        // @todo #2977109 - Calculate oauth scopes required.

        $api_path[$method] = array_filter($path_method);
      }
      // Each path contains the "base path" from a OpenAPI perspective.
      $path = str_replace($this->getJsonApiBase(), '', $route->getPath());
      $api_paths[$path] = NestedArray::mergeDeep(empty($api_paths[$path]) ? [] : $api_paths[$path], $api_path);
    }
    return $api_paths;
  }

  /**
   * Gets the JSON API routes.
   *
   * @return \Symfony\Component\Routing\Route[]
   *   The routes.
   */
  protected function getJsonApiRoutes() {
    $all_routes = $this->routingProvider->getAllRoutes();
    $jsonapi_routes = [];
    $jsonapi_base_path = $this->getJsonApiBase();
    /** @var \Symfony\Component\Routing\Route $route */
    foreach ($all_routes as $route_name => $route) {
      $is_jsonapi = $route->getDefault(JsonApiRoutes::JSON_API_ROUTE_FLAG_KEY);
      $is_entry_point = $route->getPath() === $jsonapi_base_path;
      if (!$is_jsonapi || $is_entry_point) {
        continue;
      }
      $jsonapi_routes[$route_name] = $route;
    }
    return $jsonapi_routes;
  }

  /**
   * Gets description of a method on a route.
   *
   * @param \Symfony\Component\Routing\Route $route
   *   The route.
   * @param string $route_name
   *   The route name.
   * @param string $method
   *   The method.
   *
   * @return string
   *   The method summary.
   */
  protected function getRouteMethodSummary(Route $route, $route_name, $method) {
    $resource_type = $this->getResourceTypeFromRoute($route_name, $route);
    $entity_type_id = $resource_type->getEntityTypeId();
    $bundle = $resource_type->getBundle();
    $tag = $this->getBundleTag($entity_type_id, $bundle);
    $route_type = $this->getRouteTypeByName($route_name);
    if (in_array($route_type, ['related', 'relationship'])) {
      $target_resource_type = $this->relatedResourceType($route_name, $route);
      $target_tag = $this->getBundleTag(
        $target_resource_type->getEntityTypeId(),
        $target_resource_type->getBundle()
      );
      return $this->t('@route_type: @fieldName (@targetType)', [
        '@route_type' => ucfirst($route_type),
        '@fieldName' => explode('.', $route_name)[2],
        '@targetType' => $target_tag,
        '@tag' => $tag,
      ]);
    }
    if ($route_type === 'collection') {
      if ($method === 'get') {
        return $this->t('List (@tag)', ['@tag' => $tag]);
      }
      if ($method === 'post') {
        return $this->t('Create (@tag)', ['@tag' => $tag]);
      }
    }
    if ($route_type === 'individual') {
      if ($method === 'get') {
        return $this->t('View (@tag)', ['@tag' => $tag]);
      }
      if ($method === 'patch') {
        return $this->t('Update (@tag)', ['@tag' => $tag]);
      }
      if ($method === 'delete') {
        return $this->t('Remove (@tag)', ['@tag' => $tag]);
      }
    }
    return $this->t('@route_type @method', [
      '@route_type' => ucfirst($route_type),
      '@method' => strtoupper($method),
    ]);
  }

  /**
   * Gets description of a method on a route.
   *
   * @param \Symfony\Component\Routing\Route $route
   *   The route.
   * @param string $route_name
   *   The route name.
   * @param string $method
   *   The method.
   * @param string $resource_type_name
   *   The resource type name
   *
   * @return string
   *   The method description.
   */
  protected function getRouteMethodDescription($route, $route_name, $method, $resource_type_name) {
    $route_type = $this->getRouteTypeByName($route_name);
    // @todo Make this support other methods.
    if (!$route_type || $method !== 'get') {
      return NULL;
    }
    if ($route_type === 'collection') {
      $message = '%link_co for the @name resource type. Collections are a list';
      $message .= ' of %link_ro for a particular resource type. In the JSON ';
      $message .= 'API module for Drupal all collections are homogeneous, ';
      $message .= 'which means that all the items in a collection are of the ';
      $message .= 'same type.';
      return $this->t($message, [
        '%link_co' => Link::fromTextAndUrl(
          $this->t('Collection endpoint'),
          Url::fromUri('http://jsonapi.org/format/#fetching')
        )->toString(),
        '@name' => $resource_type_name,
        '%link_ro' => Link::fromTextAndUrl(
          $this->t('resource objects'),
          Url::fromUri('http://jsonapi.org/format/#document-resource-objects')
        )->toString(),
      ]);
    }
    elseif ($route_type === 'individual') {
      $message = '%link_in for the @name resource type. The individual ';
      $message .= 'endpoint contains a %link_ro with the data for a particular';
      $message .= ' resource or entity.';
      return $this->t($message, [
        '%link_in' => Link::fromTextAndUrl(
          $this->t('Individual endpoint'),
          Url::fromUri('http://jsonapi.org/format/#fetching')
        )->toString(),
        '@name' => $resource_type_name,
        '%link_ro' => Link::fromTextAndUrl(
          $this->t('resource object'),
          Url::fromUri('http://jsonapi.org/format/#document-resource-objects')
        )->toString(),
      ]);
    }
    elseif ($route_type === 'related') {
      $message = '%link_related for the @target_name resource type through the';
      $message .= ' %field_name relationship. The related endpoint contains a ';
      $message .= '%link_ro with the data for a particular related resource or';
      $message .= ' entity.';
      $target_resource_type = $this->relatedResourceType($route_name, $route);
      return $this->t($message, [
        '%link_related' => Link::fromTextAndUrl(
          $this->t('Related endpoint'),
          Url::fromUri('http://jsonapi.org/format/#fetching')
        )->toString(),
        '@target_name' => $target_resource_type->getTypeName(),
        '%field_name' => explode('.', $route_name)[2],
        '%link_ro' => Link::fromTextAndUrl(
          $this->t('resource object'),
          Url::fromUri('http://jsonapi.org/format/#document-resource-objects')
        )->toString(),
      ]);
    }
    elseif ($route_type === 'relationship') {
      $message = '%link_rel for the @target_name resource type through the';
      $message .= ' %field_name relationship. The relationship endpoint ';
      $message .= 'contains a %link_ri with the data for a particular ';
      $message .= 'relationship.';
      $target_resource_type = $this->relatedResourceType($route_name, $route);
      return $this->t($message, [
        '%link_rel' => Link::fromTextAndUrl(
          $this->t('Relationship endpoint'),
          Url::fromUri('https://jsonapi.org/format/#fetching-relationships')
        )->toString(),
        '@target_name' => $target_resource_type->getTypeName(),
        '%field_name' => explode('.', $route_name)[2],
        '%link_ri' => Link::fromTextAndUrl(
          $this->t('resource identifier object'),
          Url::fromUri('https://jsonapi.org/format/#document-resource-identifier-objects')
        )->toString(),
      ]);
    }
    return NULL;
  }

  /**
   * Gets the route from the name if possible.
   *
   * @param string $route_name
   *   The route name.
   *
   * @return string
   *   The route type.
   */
  protected function getRouteTypeByName($route_name) {
    if (strpos($route_name, '.related') !== FALSE) {
      return 'related';
    }
    if (strpos($route_name, '.relationship') !== FALSE) {
      return 'relationship';
    }
    $route_name_parts = explode('.', $route_name);
    return isset($route_name_parts[2]) ? $route_name_parts[2] : '';
  }

  /**
   * Gets the related Resource Type.
   *
   * @param string $route_name
   *   The JSON API route name for which the ResourceType is wanted.
   * @param \Symfony\Component\Routing\Route $route
   *   The JSON API route for which the ResourceType is wanted.
   *
   * @return \Drupal\jsonapi\ResourceType\ResourceType|null
   *   Returns the ResourceType for the related JSON API resource.
   */
  protected function relatedResourceType($route_name, $route) {
    if (!in_array(
      $this->getRouteTypeByName($route_name),
      ['related', 'relationship'])
    ) {
      return NULL;
    }
    $resource_type = $this->getResourceTypeFromRoute($route_name, $route);
    $field_name = explode('.', $route_name)[2];
    $target_resource_type = current($resource_type->getRelatableResourceTypesByField($field_name));
    assert(is_a($target_resource_type, ResourceType::class));
    return $target_resource_type;
  }

  /**
   * Get the parameters array for a method on a route.
   *
   * @param \Symfony\Component\Routing\Route $route
   *   The route.
   * @param string $route_name
   *   The route name.
   * @param \Drupal\jsonapi\ResourceType\ResourceType $resource_type
   *   The JSON API resource type.
   * @param string $method
   *   The HTTP method.
   *
   * @return array
   *   The parameters.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function getMethodParameters(Route $route, $route_name, ResourceType $resource_type, $method) {
    $parameters = [];
    if ($method === 'get' && $resource_type->isVersionable()) {
      $parameters[] = [
        'name' => 'resourceVersion',
        'in' => 'query',
        'schema' => [
          'type' => 'string',
        ],
        'required' => FALSE,
        'description' => $this->t(
          'The JSON:API module exposes entity revisions as resource versions. @link.',
          [
            '@link' => Link::fromTextAndUrl(
              'Learn more in the documentation',
              Url::fromUri('https://www.drupal.org/docs/8/modules/jsonapi/revisions')
            )->toString(),
          ]
        ),
      ];
    }
    $entity_type_id = $resource_type->getEntityTypeId();
    $bundle_name = $resource_type->getBundle();
    $option_parameters = $route->getOption('parameters');
    if (!empty($option_parameters) && $filtered_parameters = $this->filterParameters($option_parameters)) {
      foreach ($filtered_parameters as $parameter_name => $parameter_info) {
        $parameter = [
          'name' => $parameter_name,
          'required' => TRUE,
          'in' => 'path',
        ];
        if ($parameter_info['converter'] === static::JSON_API_UUID_CONVERTER) {
          $parameter['schema'] = [
            'type' => 'string',
            'format' => 'uuid',
          ];
          $parameter['description'] = $this->t('The uuid of the @entity @bundle',
            [
              '@entity' => $entity_type_id,
              '@bundle' => $bundle_name,
            ]
          );
        }
        $parameters[] = $parameter;
      }

      if ($this->jsonApiPathHasRelated($route->getPath())) {
        $parameters[] = [
          'name' => 'related',
          'required' => TRUE,
          'in' => 'path',
          'schema' => [
            'type' => 'string',
          ],
          'description' => $this->t('The relationship field name'),
        ];
      }
    }
    $route_type = $this->getRouteTypeByName($route_name);
    if ($method == 'get' && $route_type === 'collection' && $resource_type->isLocatable()) {
      // If no route parameters and GET then this is collection route.
      // @todo Add descriptions or link to documentation.
      $parameters[] = [
        'name' => 'filter',
        'in' => 'query',
        'schema' => [
          // @todo Is this correct? It's a URL representation of an array.
          'type' => 'array',
        ],
        'required' => FALSE,
        'description' => $this->t('The JSON:API module has some of the most robust and feature-rich filtering features around. All of that power comes with a bit of a learning curve though. @link.', [
            '@link' => Link::fromTextAndUrl(
              'Learn more in the documentation',
              Url::fromUri('https://www.drupal.org/docs/8/modules/jsonapi/filtering')
            )->toString(),
          ]
        ),
      ];
      $parameters[] = [
        'name' => 'sort',
        'in' => 'query',
        'schema' => [
          // @todo Is this correct? It's a URL representation of an array.
          'type' => 'array',
        ],        'required' => FALSE,
        'description' => $this->t('The JSON:API module allows you to sort collections based on properties in the resource or in nested resources. @link.', [
            '@link' => Link::fromTextAndUrl(
              'Learn more in the documentation',
              Url::fromUri('https://www.drupal.org/docs/8/modules/jsonapi/sorting')
            )->toString(),
          ]
        ),
      ];
      $parameters[] = [
        'name' => 'page',
        'in' => 'query',
        'schema' => [
          // @todo Is this correct? It's a URL representation of an array.
          'type' => 'array',
        ],
        'required' => FALSE,
        'description' => $this->t('Pagination can be a deceptively complex topic. It\'s easy to fall into traps and not follow best-practices. @link.', [
            '@link' => Link::fromTextAndUrl(
              'Learn more in the documentation',
              Url::fromUri('https://www.drupal.org/docs/8/modules/jsonapi/pagination')
            )->toString(),
          ]
        ),
      ];
      $parameters[] = [
        'name' => 'include',
        'in' => 'query',
        'schema' => [
          'type' => 'string',
        ],
        'required' => FALSE,
        'description' => $this->t('Embed related entities in the response. For example: use a query string like <code>?include=comments.author</code> to include all the entities referenced by <code>comments</code> and all the entities referenced by <code>author</code> on those entities!. @link.', [
            '@link' => Link::fromTextAndUrl(
              'Learn more in the documentation',
              Url::fromUri('https://www.drupal.org/docs/8/modules/jsonapi/includes')
            )->toString(),
          ]
        ),
      ];
    }
    return $parameters;
  }

  /**
   * Get the requestBody object for a method.
   *
   * @param $resource_type
   *   Resource type.
   * @param $method
   *   Method.
   * @param $route_name
   *   Route name.
   * @param Route|NULL $route
   *   Route.
   *
   * @return array
   *   Request body object data.
   */
  protected function getRequestBody(ResourceType $resource_type, $method, $route_name, Route $route = NULL): array {
    $route_type = $this->getRouteTypeByName($route_name);
    $body_entity_type_id = $resource_type->getEntityTypeId();
    $body_bundle_name = $resource_type->getBundle();
    if (in_array($route_type, ['related', 'relationship'])) {
      $target_resource_type = $this->relatedResourceType($route_name, $route);
      $body_entity_type_id = $target_resource_type->getEntityTypeId();
      $body_bundle_name = $target_resource_type->getBundle();
    }
    if ($route_type === 'relationship') {
      $is_multiple = $this->isToManyRelationship($route_name, $resource_type);
      // Relationships are completely different.
      return [
        'description' => $this->t('The resource identifier object'),
        'required' => TRUE,
        'content' => ['application/vnd.api+json' => ['schema' => static::buildRelationshipSchema($is_multiple, $target_resource_type->getTypeName())]],
      ];
    }
    elseif ($method === 'patch') {
      // @todo Establish a subset of fields, no meta, etc.
      return [];
    }
    else {
      return [
        'description' => $this->t('The %label object', ['%label' => $body_entity_type_id]),
        'required' => TRUE,
        'content' => ['application/vnd.api+json' => ['schema' => [
          '$ref' => $this->getDefinitionReference($body_entity_type_id, $body_bundle_name),
        ]]],
      ];
    }
  }

  /**
   * Get schema for a top-level data object.
   *
   * We _could_ simply get the schema for the top-level document here, however
   * that would mean we ship separate schemas for collections and individual
   * resources, which is a lot of boilerplate considering the only difference
   * is whether the data element is an array or a single member. The code here
   * is initially lifted from JsonApiDocumentTopLevelNormalizer, but is
   * refined to allow re-usability of the contained resource type's schema.
   *
   * @param TopLevelDataInterface $object
   *   Top-level data supported for schema generation.
   *
   * @return array
   *   Schema for the top-level object.
   */
  protected function getTopLevelSchema(TopLevelDataInterface $object): array {
    $schema = [
      'allOf' => [
        ['$ref' => JsonApiSpec::SUPPORTED_SPECIFICATION_JSON_SCHEMA],
      ],
    ];

    // Top-level JSON:API documents may contain top-level data or an error
    // collection.
    $data = $object->getData();
    // Relationship data - "resource identifier object(s)"
    if ($data instanceof RelationshipData) {
      if ($data->getCardinality() === 1) {
        $schema['properties']['data'] = [
          '$ref' => JsonApiSpec::SUPPORTED_SPECIFICATION_JSON_SCHEMA . '#/definitions/relationship',
        ];
      }
      else {
        $schema['properties']['data'] = [
          'type' => 'array',
          'items' => [
            '$ref' => JsonApiSpec::SUPPORTED_SPECIFICATION_JSON_SCHEMA . '#/definitions/relationship',
          ],
          'unevaluatedItems' => FALSE,
        ];
        if ($data->getCardinality() !== FieldStorageDefinitionInterface::CARDINALITY_UNLIMITED) {
          $schema['properties']['data']['maxContains'] = $data->getCardinality();
        }
        // We can't do a minContains with the data available in this context.
      }
    }
    if ($data instanceof ResourceObjectData) {
      // @todo Need to handle cross-bundles here.
      $resource_object = $data->getData()->getIterator()->current();
      assert($resource_object instanceof ResourceObject);
      $resource_object_schema = $schema['properties']['data'] = [
        '$ref' => $this->getDefinitionReference(
          $resource_object->getResourceType()->getEntityTypeId(),
          $resource_object->getResourceType()->getBundle()
        )
      ];
      if ($data->getCardinality() === 1) {
        assert($data->count() === 1);
        $schema['properties']['data'] = $resource_object_schema;
      }
      else {
        $schema['properties']['data'] = [
          'type' => 'array',
          'items' => $resource_object_schema,
        ];
        if ($data->getCardinality() !== FieldStorageDefinitionInterface::CARDINALITY_UNLIMITED) {
          $schema['properties']['data']['maxContains'] = $data->getCardinality();
        }
      }
    }
    return $schema;
  }

  protected function getEntityResponsesJsonApi($resource_type, $method, $route_name, Route $route = NULL) {
    $route_type = $this->getRouteTypeByName($route_name);
    if ($route_type === 'collection') {
      if ($method === 'get') {
        $schema_response = ['content' => ['application/vnd.api+json' => ['schema' => $this->getTopLevelSchema($this->getStubResourceObjectData($resource_type, -1))]]];
        $responses['200'] = [
            'description' => 'successful operation',
          ] + $schema_response;
        return $responses;
      }
    }
    elseif (in_array($route_type, ['relationship', 'related'])) {
      $resource_type = $this->getResourceTypeFromRoute($route_name, $route);
      $target_resource_type = $this->relatedResourceType($route_name, $route);
      $is_multiple = $this->isToManyRelationship($route_name, $resource_type);
      if ($route_type === 'relationship') {
        $schema = static::buildRelationshipSchema($is_multiple, $target_resource_type->getTypeName());
        if ($method === 'get') {
          return [
            200 => [
              'description' => 'successful operation',
              'content' => ['application/vnd.api+json' => ['schema' => $schema]],
            ],
          ];
        }
        elseif ($method === 'post') {
          return [201 => ['description' => 'created', 'content' => ['application/vnd.api+json' => ['schema' => $schema]]]];
        }
        elseif ($method === 'patch') {
          return [
            200 => [
              'description' => 'successful operation',
              'content' => ['application/vnd.api+json' => ['schema' => $schema]],
            ],
          ];
        }
        elseif ($method === 'delete') {
          return [204 => ['description' => 'no content']];
        }
      }
      else {
        // Fake a route name that will yield the expected results for the related
        // responses.
        $target_route_name = Routes::getRouteName($target_resource_type, $is_multiple ? 'collection' : 'individual');
        return $this->getEntityResponsesJsonApi(
          $resource_type,
          $method,
          $target_route_name
        );
      }
    }
    else {
      $responses = [];

      $schema_response = ['content' => ['application/vnd.api+json' => ['schema' => $this->getTopLevelSchema($this->getStubResourceObjectData($resource_type))]]];

      switch ($method) {
        case 'get':
          $responses['200'] = [
              'description' => 'successful operation',
            ] + $schema_response;
          break;

        case 'post':
          unset($responses['200']);
          $responses['201'] = [
              'description' => 'Entity created',
            ] + $schema_response;
          break;

        case 'delete':
          unset($responses['200']);
          $responses['201'] = [
            'description' => 'Entity deleted',
          ];
          break;
      }
      return $responses;
    }
    return [];
  }

  /**
   * Load a resource type and also mark it to be included in the schemas object.
   *
   * @param string $entity_type_id
   *   Entity Type ID.
   * @param string|null $bundle
   *   Bundle name.
   *
   * @return ResourceType
   *   The resource type.
   */
  protected function loadResourceType(string $entity_type_id, ?string $bundle = NULL): ResourceType {
    $definition_key = $this->normalizeResourceTypeId($entity_type_id, $bundle);
    if (empty($this->resourceTypes[$definition_key])) {
      $this->resourceTypes[$definition_key] = $this->resourceTypeRepository->get($entity_type_id, $bundle);
    }
    return $this->resourceTypes[$definition_key];
  }

  protected function storeResourceType(ResourceType $resource_type): void {
    $id = $this->normalizeResourceTypeId($resource_type->getEntityTypeId(), $resource_type->getBundle());
    if (empty($this->resourceTypes[$id])) {
      $this->resourceTypes[$id] = $resource_type;
    }
  }

  protected function normalizeResourceTypeId(string $entity_type_id, ?string $bundle = NULL): string {
    if (!$bundle) {
      $bundle = $entity_type_id;
    }
    return "{$entity_type_id}-{$bundle}";
  }

  /**
   * {@inheritdoc}
   */
  protected function getJsonSchema(string $entity_type_id, ?string $bundle = NULL): array {
    // JSON:API module specific note - this gets resource type schemas, only;
    // top-level schema is handled in its own method.
    $additional = [];
    $stub_resource = $this->getStubResourceObjectData($this->loadResourceType($entity_type_id, $bundle))
      ->getData()
      ->getIterator()
      ->current();
    if (($example_normalization = $this->serializer->normalize($stub_resource, 'api_json', ['account' => NULL])) && $example_normalization instanceof CacheableNormalization) {
      $additional['examples'][] = $example_normalization->getNormalization();
    }
    return $this->serializer->normalize(
      $stub_resource,
      'json_schema'
    ) + $additional;
  }

  protected function getStubResourceObjectData(ResourceType $resource_type, int $cardinality = 1): ResourceObjectData {
    if (!$this->stubDataStorage->offsetExists($resource_type)) {
      $entity_storage = $this->entityTypeManager->getStorage($resource_type->getEntityTypeId());
      $stub_entity = $entity_storage instanceof ContentEntityStorageInterface
        ? $entity_storage->createWithSampleValues($resource_type->getBundle())
        : $entity_storage->create();
      $this->stubDataStorage->offsetSet(
        $resource_type,
        $stub_entity
      );
    }
    else {
      $stub_entity = $this->stubDataStorage->offsetGet($resource_type);
    }
    return new ResourceObjectData([ResourceObject::createFromEntity(
      $resource_type,
      $stub_entity
    )], $cardinality);
  }

  /**
   * {@inheritdoc}
   */
  public function getTags() {
    $tags = [];
    foreach ($this->entityTypeManager->getDefinitions() as $entity_type) {
      if ($bundle_type_id = $entity_type->getBundleEntityType()) {
        $bundle_storage = $this->entityTypeManager->getStorage($bundle_type_id);
        $bundles = $bundle_storage->loadMultiple();
        foreach ($bundles as $bundle_name => $bundle) {
          if (!$this->includeEntityTypeBundle($entity_type->id(), $bundle_name)) {
            continue;
          }
          $description = $this->t("@bundle_label @bundle of type @entity_type.",
            [
              '@bundle_label' => $entity_type->getBundleLabel(),
              '@bundle' => $bundle->label(),
              '@entity_type' => $entity_type->getLabel(),
            ]
          );
          $tag = [
            'name' => $this->getBundleTag($entity_type->id(), $bundle->id()),
            'description' => $description,
            'x-entity-type' => $entity_type->id(),
            'x-definition' => [
              '$ref' => $this->getDefinitionReference($entity_type->id(), $bundle_name),
            ],
          ];
          if (method_exists($bundle, 'getDescription')) {
            $tag['description'] .= ' ' . $bundle->getDescription();
          }
          $tags[] = $tag;
        }
      }
      else {
        if (!$this->includeEntityTypeBundle($entity_type->id())) {
          continue;
        }
        $tag = [
          'name' => $this->getBundleTag($entity_type->id()),
        ];
        if ($entity_type instanceof ConfigEntityTypeInterface) {
          $tag['description'] = $this->t('Configuration entity @entity_type', ['@entity_type' => $entity_type->getLabel()]);
        }
        $tags[] = $tag;
      }
    }
    return $tags;
  }

  /**
   * Get the tag to use for a bundle.
   *
   * @param string $entity_type_id
   *   The entity type.
   * @param string $bundle_name
   *   The entity type.
   *
   * @return string
   *   The bundle tag.
   */
  protected function getBundleTag($entity_type_id, $bundle_name = NULL) {
    static $tags = [];
    if (!isset($tags[$entity_type_id][$bundle_name])) {
      $entity_type = $this->entityTypeManager->getDefinition($entity_type_id);
      $tag = $entity_type->getLabel();
      if ($bundle_name && $bundle_type_id = $entity_type->getBundleEntityType()) {
        $bundle_entity = $this->entityTypeManager->getStorage($bundle_type_id)
          ->load($bundle_name);
        $tag .= ' - ' . $bundle_entity->label();
      }
      $tags[$entity_type_id][$bundle_name] = $tag;
    }
    return $tags[$entity_type_id][$bundle_name];
  }

  /**
   * {@inheritdoc}
   */
  public function getApiName() {
    return $this->t('JSON API');
  }

  /**
   * {@inheritdoc}
   */
  protected function getApiDescription() {
    return $this->t('This is a JSON API compliant implementation');
  }

  /**
   * Gets a Resource Type for a given route.
   *
   * @todo This should support cross bundles module/mixed resource types.
   *
   * @param string $route_name
   *   The JSON API route name for which the ResourceType is wanted.
   * @param \Symfony\Component\Routing\Route $route
   *   The JSON API route for which the ResourceType is wanted.
   *
   * @return \Drupal\jsonapi\ResourceType\ResourceType|null
   *   Returns the ResourceType for the given JSON API route, NULL if the route
   *   parameter could not be upcasted.
   */
  protected function getResourceTypeFromRoute($route_name, Route $route) {
    $parameters[RouteObjectInterface::ROUTE_NAME] = $route_name;
    $parameters[RouteObjectInterface::ROUTE_OBJECT] = $route;
    $upcasted_parameters = $this->paramConverterManager->convert($parameters + $route->getDefaults());
    return $upcasted_parameters[JsonApiRoutes::RESOURCE_TYPE_KEY] ?? NULL;
  }

  /**
   * Filters an associative array by key on a set of parameter.
   *
   * @param array $parameters
   *   Associative array that is going to be filtered.
   *
   * @return array
   *   Returns the filtered associative array.
   */
  protected function filterParameters(array $parameters) {
    foreach (static::$PARAMETERS_FILTER_LIST as $filter) {
      if (array_key_exists($filter, $parameters)) {
        unset($parameters[$filter]);
      }
    }
    return $parameters;
  }

  /**
   * Checks if a JSON API Path has {related}.
   *
   * @todo remove once https://www.drupal.org/project/jsonapi/issues/2953346
   * is done on JSON API Project.
   *
   * @param string $path
   *   The path.
   *
   * @return bool
   *   TRUE if path contains {related}, FALSE otherwise
   */
  protected function jsonApiPathHasRelated($path) {
    return strpos($path, '{related}') !== FALSE;
  }

  /**
   * Builds the relationship schema.
   *
   * @param bool $is_multiple
   *   Indicates if the relationship is to-many.
   * @param string $resource_type_name
   *   The resource type for the relationship.
   *
   * @return array
   *   The schema definition.
   *
   * @todo: build this once and use '$ref' when necessary.
   */
  protected static function buildRelationshipSchema($is_multiple, $resource_type_name = NULL) {
    $linkage_schema = [
      'description' => 'The "type" and "id" to non-empty members.',
      'type' => 'object',
      'required' => ['type', 'id'],
      'properties' => [
        'type' => ['title' => t('Resource type name'), 'type' => 'string'],
        'id' => ['title' => t('Resource ID'), 'type' => 'string', 'format' => 'uuid'],
        'meta' => [
          'description' => 'Non-standard meta-information that can not be represented as an attribute or relationship.',
          'type' => 'object',
          'additionalProperties' => TRUE,
          'properties' => (object) [],
        ],
      ],
      'additionalProperties' => FALSE,
    ];
    if ($resource_type_name) {
      $linkage_schema['properties']['type']['enum'] = [$resource_type_name];
    }
    return [
      'type' => 'object',
      'properties' => [
        'data' => $is_multiple
          ? [
            'description' => 'An array of objects each containing \"type\" and \"id\" members for to-many relationships.',
            'type' => 'array',
            'items' => $linkage_schema,
            'uniqueItems' => TRUE,
          ]
          : $linkage_schema
      ],
    ];
  }

  /**
   * Checks if the relationship is to-many.
   *
   * @param string $route_name
   *   The route name.
   * @param \Drupal\jsonapi\ResourceType\ResourceType $resource_type
   *   The resource type.
   *
   * @return bool
   *   Indicates if the relationship is multiple cardinality.
   */
  protected function isToManyRelationship($route_name, ResourceType $resource_type) {
    $public_field_name = explode('.', $route_name)[2];
    $internal_field_name = $resource_type->getInternalName($public_field_name);
    $field_definitions = $this->fieldManager
      ->getFieldDefinitions($resource_type->getEntityTypeId(), $resource_type->getBundle());
    return $field_definitions[$internal_field_name]->getFieldStorageDefinition()->getCardinality() !== 1;
  }

}
